#!/usr/bin/python

# Copyright (c) 2012, Hallgeir Lien <hallgeir.lien@gmail.com> 
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met: 

# Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.

# Neither the name of the <ORGANIZATION> nor the names of its contributors may
#   be used to endorse or promote products derived from this software without
#   specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import sqlite3, subprocess, sys, re, os, time
from math import *
from datetime import datetime

infinity = 1e127
flags = {"verbose": False}

def median(a):
    return sorted(a)[len(a)/2]

def mean(a):
    m = 0
    for v in a:
        m += v
    return m/float(len(a))

def stddev(a, m = None):
    if len(a) <= 1: return 0
    if m == None:
        m = mean(a)
    out = 0
    for v in a:
        out += (v-m)**2

    return sqrt(out/float(len(a)))

def log_write(logfiles, text, verbose = False):
    if logfiles == None: 
        return

    for l in logfiles:
        if (l["verbose"] == True or flags["verbose"]) or verbose == False:
            l["stream"].write(datetime.now().strftime("%Y-%m-%d %H:%M:%S") + " [Info]  " + text + "\n");

def error_write(logfiles, text):
    if logfiles == None: 
        return

    for l in logfiles:
        l.write(str(datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + " [ERROR] " + text + "\n");

class Test:
    def __init__(self):
        self.name = ""
        self.executable = ""
        self.path = "./"
        self.makeflags = ""
        self.inputs = []

    def __str__(self):
        return str(self.__dict__)

class Input:
    def __init__(self, input_params, _file):
        self.file = _file   #Set to filename if output is already stored in file
        self.input_params = input_params
        self.repcount = 0 #Used to know how many reps we have done

    def __str__(self):
        return str(self.__dict__)

class Result:
    def __init__(self, starttime, results = {}):
        self.results = results.copy()
        self.starttime = starttime

    def set(self, key, value):
        self.results[key] = value
    
    def get(self, key):
        return self.results.get(key, None)
    
    def keys(self):
        return self.results.keys()

    def __str__(self):
        return str(self.__dict__)

class ResultSet:
    def __init__(self, _input):
        self.results = []
        self.stats = {}
        self.input = _input
        for s in ["mean", "median", "stddev", "range_min", "range_max"]:
            self.stats[s] = {}
    
    def add(self, result):
        # Add the result
        self.results.append(result)
        self.update_stats()
 
    def add_set(self, other):
        if self.input != other.input:
            raise ValueError("Cannot join two resultsets with different input.")

        self.results += other.results
        self.update_stats()

    def update_stats(self):
        # Update statistics
        keys = self.keys()

        for k in keys:
            self.stats["mean"][k] = []
            self.stats["median"][k] = []
            self.stats["stddev"][k] = []
            self.stats["range_min"][k] = infinity
            self.stats["range_max"][k] = -infinity

        for r in self.results:
            for k in keys:
                v = r.get(k)
                self.stats["mean"][k] += [v]
                self.stats["median"][k] += [v]
                self.stats["stddev"][k] += [v]
                self.stats["range_min"][k] = min(v, self.stats["range_min"][k])
                self.stats["range_max"][k] = max(v, self.stats["range_max"][k])

        for k in keys:
            self.stats["mean"][k] = mean(self.stats["mean"][k])
            self.stats["median"][k] = median(self.stats["median"][k])
            self.stats["stddev"][k] = stddev(self.stats["stddev"][k], m = self.stats["mean"][k])

    def keys(self):
        if len(self.results) > 0:
            return self.results[0].keys()

        return []

    def median(self, key):
        return self.stats["median"].get(key, None)

    def mean(self, key):
        return self.stats["mean"].get(key, None)

    def stddev(self, key):
        return self.stats["stddev"].get(key, None)

    def range_min(self, key):
        return self.stats["range_min"].get(key, None)

    def range_max(self, key):
        return self.stats["range_max"].get(key, None)

class ResultValue:
    def __init__(self, type, regex=None):
        self.value = None
        self.type = type
        self.regex = regex

def run_tests(test, logfiles, errorfiles):
    results = []
    resultsets = {}
    suite = None

    if test.executable == "elib_test.x":
        suite = Elib_Suite()
    elif test.executable == "cg":
        suite = CSX_Suite()

    log_write(logfiles, "Test \"%s\" starts at %s" % (test.name, str(datetime.now())))

    oldinput = None
 
    remaining_input = test.inputs[:]

    while len(remaining_input) > 0:
        for i in remaining_input:
            the_suite = suite
            if i.input_params.get("valgrind", "off") == "on":
                the_suite = Valgrind_Suite(suite)
    
            if the_suite.must_make(i, oldinput, test):
                the_suite.make(test, i, logfiles, errorfiles)
                oldinput = i
    
            r = the_suite.test(test, i, os.path.join(test.path, test.executable), logfiles, errorfiles)
            r.input = i
            if i not in resultsets.keys():
                resultsets[i] = r
            else:
                resultsets[i].add_set(r)

        for ii in xrange(len(remaining_input)-1, -1, -1):
            i = remaining_input[ii]
            if i.repcount >= int(i.input_params.get("reps", 5)):
                remaining_input.pop(ii)

    log_write(logfiles, "Test ends at " + str(datetime.now()))

    return resultsets.values()

def make(path, makeopts, logfiles, errorfiles):
    log_write(logfiles, "Building...")
    log_write(logfiles, "Options: " + " ".join(makeopts))

    #clean
    args = ("make -C %s clean" % path).split() 
    logstring = ' '.join(args) + " &&"
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output,errors = p.communicate()

    if errors != None and errors.strip() != "":
        error_write(errorfiles, "Make error: " + errors + ". Command: " + " ".join(args) + "\n")

    args = ("make -j4 ").split() + makeopts.split() + ("-C " + path).split()
    log_write(logfiles, logstring + ' '.join(args))
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output,errors = p.communicate()

    if errors != None and errors.strip() != "":
        error_write(errorfiles, "Make error: " + errors + ". Command: " + " ".join(args) + "\n")

# Parse environment variable string
def parse_envlist(env):
    items=env.strip().split()
    env_out = {}
    for i in items:
        kv = i.split("=", 1)
        env_out[kv[0]] = kv[1] if len(kv) > 1 else ""

    return env_out

def test_common(test, _input, result_values, executable, env, logfiles, errorfiles, result_in_stderr = False):
    starttime_best = datetime.now()
    results = {}

    # Set CPU frequency
    cpufreq = _input.input_params.get("cpufreq", None)
    if cpufreq is not None:
        log_write(logfiles, "Setting CPU frequency to %d MHz..." % int(cpufreq))
        p = subprocess.Popen(["./set_cpufreq", str(int(cpufreq))], stdout=subprocess.PIPE, stderr=subprocess.PIPE);
        output,error=p.communicate()
        if error: print error

    if _input.file == None:
        args = executable.split(" ")
        args += _input.input_params.get("argv", "").split()
    else:
        args = ("cat %s " % (_input.file)).split()

    # Update the environment variables
    envs = os.environ.copy()
    env_extras = parse_envlist(_input.input_params.get("env", ""))

    for k,v in env.items():
        envs[k] = v

    for k,v in env_extras.items():
        envs[k] = v

    log_write(logfiles, "Executing %s. Environment variables: %s %s. Iterations: %d" % (' '.join(args), str(env), str(env_extras), int(_input.input_params.get("reps", 5))))

    # JMCG

    proctemp = _input.input_params.get("proctemp", None)
    yokogawa_log = _input.input_params.get("yokogawa_log", None)
    
    if ((yokogawa_log is not None) and (yokogawa_log != "0")):
        log_write(logfiles, "Starting Intel Energy Server on remove server")
        subprocess.Popen(["./start_yoko_log", "cg_%s_%s_mt-%s_xf-%s_st-%s" % (_input.input_params.get("testset", None), _input.input_params.get("matrix", None), _input.input_params.get("mt_conf", None), _input.input_params.get("xform_conf", None), _input.input_params.get("enable_stat_preprocessing", None))], stdout=subprocess.PIPE, stderr=subprocess.PIPE);

        while (os.path.isfile("./sync_ready") == False):
            time.sleep(1)
    
        subprocess.Popen("rm ./sync_ready", shell=True)

    if proctemp is not None:
        log_write(logfiles, "Setting CPU temperature to %d C..." % int(proctemp)) 
        p = subprocess.Popen(["./set_temp", proctemp], stdout=subprocess.PIPE, stderr=subprocess.PIPE);
        output,error=p.communicate()
        if error: print error
   # JMCG

    results = ResultSet(_input)

    for j in xrange(int(_input.input_params.get("reps", 5))):

        _input.repcount += 1;

        log_write(logfiles, "Starting iteration %d." % _input.repcount, True);

        # Sleep in order to get good yokogawa results
        if ((yokogawa_log is not None) and (yokogawa_log != "0")):
            log_write(logfiles, "Sleeping for 5 secs", True)
            time.sleep(5)
            log_write(logfiles, "Done sleepan!", True)

        result = Result(datetime.now())

        # Execute the process
        log_write(logfiles, "Program is being run...", True)
        p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=envs)
        if result_in_stderr:
            tmp,output = p.communicate()
            errors = None
        else:
            output,errors = p.communicate()

        log_write(logfiles, "Program output:\n%s" % output.strip(), verbose=True)
        log_write(logfiles, "Program output end", verbose=True)
        if errors != None and errors.strip() != "":
            error_write(errorfiles, "Iteration " + str(_input.repcount) + ": " + errors.strip())
        if p.returncode != 0:
            error_write(errorfiles, "Encountered error code %d. Exiting." % p.returncode)
            sys.exit(p.returncode)

        outstring = ""
        for k,v in result_values.items():
            regex = re.compile(v.regex)
            matches = regex.search(output)
            if matches == None or matches.lastindex == None:
                error_write(errorfiles, "No matches on regex %s. Program output: %s" % (v.regex, output));
                continue
            val = v.type(matches.group(1).replace(",",""))
            result.set(k, val)
            outstring += "%s: %s " % (k, val)

        log_write(logfiles, "Results: " + outstring)

        results.add(result)

    if ((yokogawa_log is not None) and (yokogawa_log != "0")):
        log_write(logfiles, "Sleeping for 2 secs", True)
        time.sleep(2)
        log_write(logfiles, "Done sleepan!", True)
        log_write(logfiles, "Stop Intel Energy Server on remove server")
        p = subprocess.Popen(["./stop_yoko_log"], stdout=subprocess.PIPE, stderr=subprocess.PIPE);
        output,error=p.communicate()
        if error: print error

    return results

#
# Memory behavior testing with Valgrind
#
class Valgrind_Suite:
    # Takes in a test suite which valgrind will test
    def __init__(self, suite):
        self.suite = suite

    def must_make(self, inext, iprev, test):
        return self.suite.must_make(inext, iprev, test)
    
    def make(self, test, _input, logfiles, errorfiles):
        self.suite.make(test, _input, logfiles, errorfiles)

    def arglist(self, test, _input, executable):
        args = "valgrind --tool=callgrind --callgrind-out-file=/dev/null "
        if _input.input_params.get("L1", None) != None:
            L1 = _input.input_params["L1"].split(",")
            args += "--D1=%d,%d,%d " % (int(L1[0])*1024, int(L1[1]), int(L1[2]))
        if _input.input_params.get("LL", None) != None:
            L2 = _input.input_params["LL"].split(",")
            args += "--LL=%d,%d,%d " % (int(L2[0])*1024, int(L2[1]), int(L2[2]))

        args += "--collect-atstart=no "

        args += self.suite.arglist(test, _input, executable)
        return args

    def envlist(self, test, _input):
        return self.suite.envlist(test, _input)

    def test(self, test, _input, executable, logfiles, errorfiles):
        return test_common(test, _input,  
            {"cache_data_refs": ResultValue(int, "D\s+refs:\s+(\S+)"), 
             "cache_l1_miss": ResultValue(int, "D1\s+misses:\s+(\S+)"), 
             "cache_ll_miss": ResultValue(int, "LLd\s+misses:\s+(\S+)"), 
             "cache_ll_refs": ResultValue(int, "D1\s+misses:\s+(\S+)")}, 
            self.arglist(test, _input, executable), self.envlist(test, _input), logfiles, errorfiles, True)

#
# Energy library testing functions
#
class Elib_Suite:
    # Determine if we need to remake, given the next and previous input
    def must_make(self, inext, iprev, test):
        if iprev != None:
            return False
        else:
            return True

    def make(self, test, _input, logfiles, errorfiles):
        make(test.path, test.makeflags, logfiles, errorfiles)

    def arglist(self, test, _input, executable):
        args = os.path.join(test.path, test.executable)
        return args

    def envlist(self, test, _input):
        return {}
        
    def test(self, test, _input, executable, logfiles, errorfiles):
        return test_common(test, _input,
            {"rt": ResultValue(float, "rt: (\S+)"), 
             "energy": ResultValue(float, "energy: (\S+)"), 
             "energy_pp0": ResultValue(float, "pp0: (\S+)"), 
             "energy_pp1": ResultValue(float, "pp1: (\S+)")}, 
            self.arglist(test, _input, executable), self.envlist(test, _input), logfiles, errorfiles)

#
# CSX testing function
#
class CSX_Suite:
    # Determine if we need to remake, given the next and previous input
    def must_make(self, inext, iprev, test):
        return True

    def make(self, test, _input, logfiles, errorfiles):
        return True

    def arglist(self, test, _input, executable):
        args = os.path.join(test.path, test.executable)
        if int(_input.input_params.get("symmetry", 0)) == 1:
            args = args + " -s"
        if int(_input.input_params.get("compress", 0)) == 1:
            args = args + " -x"
        args = args + " -l " + str(int(_input.input_params.get("cg_iterations", 1024)))
        args = args + " -L " + str(int(_input.input_params.get("cg_reps", 1)))
        args = args + " " + str(_input.input_params.get("matrix", None))
        
        return args

    def envlist(self, test, _input):
        if int(_input.input_params.get("enable_stat_preprocessing",0)) == 1:
            return {"XFORM_CONF": str(_input.input_params.get("xform_conf", 0)),
                    "MT_CONF": str(_input.input_params.get("mt_conf", 0)),
                    "SAMPLES": str(int(_input.input_params.get("samples", 1))),
                    "SAMPLING_PORTION": str(float(_input.input_params.get("sampling_portion", 0.01)))}
        else:
            return {"XFORM_CONF": str(_input.input_params.get("xform_conf", 0)),
                    "MT_CONF": str(_input.input_params.get("mt_conf", 0))}
        

    def test(self, test, _input, executable, logfiles, errorfiles):
        return test_common(test, _input,
            {"cg_iterations_done": ResultValue(int, "l:(\S+)"),
             "relative_distance": ResultValue(float, "rd:(\S+)"),
             "spmv_multiplication_time": ResultValue(float, "st:(\S+)"),
             "spmv_reduction_time": ResultValue(float, "rt:(\S+)"),
             "cg_execution_time": ResultValue(float, "ct:(\S+)"),
             "energy_pkg": ResultValue(float, "energy: (\S+)"),
             "energy_pp0": ResultValue(float, "energy_pp0: (\S+)"), 
             "energy_pp1": ResultValue(float, "energy_pp1: (\S+)"),
             "papi_l2_refs": ResultValue(int, "l2_refs: (\S+)"),
             "papi_l2_miss": ResultValue(int, "l2_miss: (\S+)"),
             "papi_l3_refs": ResultValue(int, "l3_refs: (\S+)"),
             "papi_l3_miss": ResultValue(int, "l3_miss: (\S+)")},
            self.arglist(test, _input, executable), self.envlist(test, _input), logfiles, errorfiles)

class DBField:
    dtype_strings = {int: "INT", str: "VARCHAR(200)", float: "REAL", datetime: "DATETIME"}

    def __init__(self, name, dtype, pk=False):
        self.name = name
        self.dtype = dtype
        self.pk = pk

    def format_rhs(self, val):
        if self.dtype in [float, int]:
            return str(val)
        else:
            return "'%s'" % str(val)

    def get_create_string(self):
        return self.name.strip() + " " + DBField.dtype_strings[self.dtype] + (" PRIMARY KEY" if self.pk else "")

    def __str__(self):
        return self.name + ": " + str(self.dtype)

def get_table_join_string(tables, join_on):
    output_tables = ""
    first = True
    for t in tables:
        if not first:
            output_tables += ", "
        output_tables += t.name
        first = False

    output_where = ""
    first = True
    first_tab = True
    for t in tables:
        if first_tab:
            first_tab = False
            continue
        for j in join_on:
            if not first:
                output_where += " AND "
            output_where += tables[0].name + "." + j.name + "=" + t.name + "." + j.name
            first = False

    return (output_tables, output_where)

class Table:
    no_value = ["valgrind", "argv", "env", "reps"] # Fields from test input scripts that should not be stored

    def __init__(self, name, fields):
        self.name = name
        self.fields = fields

    def has_field(self, key):
        return self.get_field(key) is not None

    def get_field(self, key):
        for f in self.fields:
            if f.name == key:
                return f
        return None

    def get_create_string(self):
        first = True
        output = "CREATE TABLE IF NOT EXISTS " + self.name + " ("
        for f in self.fields:
            if not first:
                output += ", "
            output += f.get_create_string()
            first = False

        output += ")"

        return output

    def get_field_list(self, exclude_list = [], alias_to_self = False):
        output = ""
        first = True
        for f in self.fields:
            if f.name in exclude_list: continue
            if not first:
                output += ", "
            output += self.name + "." + f.name + (" AS " + f.name if alias_to_self else "")
            first = False
        return output

    def get_delete_string(self, where):
        return "DELETE FROM " + self.name + " WHERE " + where

    def get_insert_string(self, values, logfiles):
        kv = values.items()

        output = "INSERT INTO %s (" % self.name
        first = True
        for k,v in kv:
            if k in Table.no_value:
                continue

            if not self.has_field(k):
                log_write(logfiles, "Key not found in table %s: %s" % (self.name, k) + ". Possible values: " + str([x.name for x in self.fields]), True)
                continue

            if not first:
                output += ", "
            output += k
            first = False
            
        output += ") VALUES ("
        first = True
        for k,v in kv:
            if k in Table.no_value:
                continue

            f = self.get_field(k)
            if f == None:
                continue

            if not first:
                output += ", "
            output += f.format_rhs(v)
            first = False
        output += ")"
    
        return output

class DB:
    input_values = [("cpufreq", int), ("proctemp", int), ("yokogawa_log", int),
                    ("symmetry", int), ("compress", int), ("cg_reps", int), ("cg_iterations", int), ("matrix", str), ("mt_conf", str), ("xform_conf", str), ("enable_stat_preprocessing", int), ("samples", int), ("sampling_portion", float)]
    result_values = [("rt", float),
                     ("energy_pkg", float), ("energy_pp0", float), ("energy_pp1", float),
#                     ("cache_data_refs", int), ("cache_l2_refs", int), ("cache_ll_refs", int), ("cache_l1_miss", int), ("cache_l2_miss", int), ("cache_ll_miss", int),
                     ("cg_iterations_done", int), ("relative_distance", float), ("spmv_multiplication_time", float), ("spmv_reduction_time", float), ("cg_execution_time", float),
                     ("papi_l2_refs", int), ("papi_l2_miss", int), ("papi_l3_refs", int), ("papi_l3_miss", int)]

    
    testset_key = [DBField("test_name", str), DBField("testset", str)]

    testset_table = Table("testset", testset_key)
    results_table = Table("result_set", testset_key + [DBField("starttime", datetime)] + [DBField(x[0], x[1]) for x in input_values] + [DBField(x[0], x[1]) for x in result_values]) 
    stats_table   = Table("stats", testset_key + [DBField(x[0], x[1]) for x in input_values] + [DBField(x[0] + "_" + y, x[1]) for x in result_values for y in ["stddev", "median", "mean", "range_min", "range_max"]])

    tables = [testset_table, results_table, stats_table]

    def __init__(self, dbfile, logfiles = []):
        self.dbfile = dbfile
        self.db = None
        self.dbc = None
        self.logfiles = logfiles

        self.create()

    def open(self):
        self.db = sqlite3.connect(self.dbfile)
        self.dbc = self.db.cursor()
    
    def close(self):
        self.dbc.close()
        self.db.close()

        self.dbc = None
        self.db = None

    # Create all tables
    def create(self):
        self.open()
        # Create all tables
        for t in DB.tables:
            self.dbc.execute(t.get_create_string())

        # Add any missing columns
        for t in DB.tables:
            for f in t.fields:
                try:
                    q = "ALTER TABLE %s ADD COLUMN %s" % (t.name, f.get_create_string())
                    self.dbc.execute(q)
                    log_write(self.logfiles, "Executing query: %s" % (q))
                except:
                    pass

        # Create views
        dq = "DROP VIEW IF EXISTS results"
        log_write(self.logfiles, "Executing query: %s" % dq, True)
        self.dbc.execute(dq)

        view_tables,view_where = get_table_join_string([DB.testset_table, DB.stats_table], [DB.testset_key[0], DB.testset_key[1]])

        q = "SELECT %s, %s FROM %s WHERE %s" % (DB.testset_table.get_field_list(alias_to_self=True), DB.stats_table.get_field_list([x.name for x in DB.testset_key], alias_to_self=True), view_tables, view_where)
        q = "CREATE VIEW results AS %s" % q
        log_write(self.logfiles, "Executing query: %s" % q, True)
        self.dbc.execute(q)

        self.dbc.execute("CREATE INDEX IF NOT EXISTS idx_stats_key ON stats (%s, %s)" % (DB.testset_key[0].name, DB.testset_key[1].name))
        self.dbc.execute("CREATE INDEX IF NOT EXISTS idx_stats_input ON stats (%s)" % (", ".join([x[0] for x in DB.input_values])))
        self.dbc.execute("CREATE INDEX IF NOT EXISTS idx_testset ON stats (%s, %s, %s)" % (DB.testset_key[0].name, DB.testset_key[1].name, ", ".join([x[0] for x in DB.input_values])))

        self.close()

    def create_indices(self):
        self.open()
        self.dbc.execute("CREATE INDEX IF NOT EXISTS idx_stats_key ON stats (%s, %s)" % (DB.testset_key[0].name, DB.testset_key[1].name))
        self.dbc.execute("CREATE INDEX IF NOT EXISTS idx_stats_input ON stats (%s)" % (", ".join([x[0] for x in DB.input_values])))
        self.dbc.execute("CREATE INDEX IF NOT EXISTS idx_testset ON stats (%s, %s, %s)" % (DB.testset_key[0].name, DB.testset_key[1].name, ", ".join([x[0] for x in DB.input_values])))
        self.close()
        

    # Delete previous results from the database and insert the new result sets
    def delete_old_results(self, test, results):
        self.open()
        deleted = {}
        for r in results:
            ts = r.input.input_params.get("testset", "default_set")
            if (test.name, ts) not in deleted.keys():
                for t in DB.tables:
                    q = t.get_delete_string("%s = '%s' and %s = '%s'" % (DB.testset_key[0].name, test.name, DB.testset_key[1].name, ts))
                    log_write(self.logfiles, "Executing query: %s" % q, True)
                    deleted[(test.name, ts)] = True
                    self.dbc.execute(q)

        inserted = {}
        for r in results:
            ts = r.input.input_params.get("testset", "default_set")
            if (test.name, ts) not in inserted.keys():
                q = DB.testset_table.get_insert_string({DB.testset_key[0].name: test.name, DB.testset_key[1].name: ts}, self.logfiles)
                log_write(self.logfiles, "Executing query: %s" % q, True)
                inserted[(test.name, ts)] = True
                self.dbc.execute(q)

        self.db.commit()
        self.close()
        
    def insert_resultset(self, test, resultset):
        self.open()

        # Insert the testset into the testset list, if neccesary

        has_stats = {}

        # Insert results 
        for r in resultset.results:
            values = dict([(DB.testset_key[0].name, test.name)] + [("starttime", r.starttime)] + resultset.input.input_params.items() + r.results.items())
            q = DB.results_table.get_insert_string(values, self.logfiles)
            log_write(self.logfiles, "Executing query: %s" % q, True)
            self.dbc.execute(q)

        # Insert statistics
        values = [(DB.testset_key[0].name, test.name)] + resultset.input.input_params.items()
        for k in resultset.keys():
            values += [(k + "_median", resultset.median(k))]
            values += [(k + "_mean", resultset.mean(k))]
            values += [(k + "_stddev", resultset.stddev(k))]
            values += [(k + "_range_min", resultset.range_min(k))]
            values += [(k + "_range_max", resultset.range_max(k))]
        
        q = DB.stats_table.get_insert_string(dict(values), self.logfiles)
        log_write(self.logfiles, "Executing query: %s" % q, True)
        self.dbc.execute(q)

        self.db.commit()
        self.close()

def print_help():
    print("Usage:")
    print("./test.py OPTIONS")
    print("Options:")
    print("-o <file>, set database output file to <file>")
    print("-i <file>, set input file to <file>")
    print("-l <file>, set logging file to <file>")
    print("-v, set verbose flag")

#def read_file(filename, level, t, testset, width, height, nthreads, vectorize):
def read_file(filename, level, t, input_params):
    inputdata = open(filename, "r")
    out = []

    for i in inputdata:
        if len(i.strip()) == 0 or i.strip()[0] == "#": continue
        inp = i.strip().split()
        inp[0] = inp[0].strip()
        if inp[0] == "test": t.name = inp[1]
        elif inp[0] == "path": t.path = inp[1]
        elif inp[0] == "exec": t.executable = inp[1]
        elif inp[0] == "makeflags": t.makeflags = " ".join(inp[1:])
        elif inp[0] == "run": out.append( Input(input_params.copy(), None) )
        elif inp[0] == "include":
            ins = read_file(''.join(inp[1:]), level+1, t, input_params)
            out += ins
        else:
            input_params[inp[0]] = i.strip()[len(inp[0]):].strip() 
    inputdata.close()

    return out

def main(args):
    output = None # Output filename
    program = None
    inputfile = "input.in"
    logfile = None

    try:
        i = 0
        while i < len(args):
            if args[i] == "-i":
                inputfile = args[i+1]
                i += 1
            elif args[i] == "-o":
                output = args[i+1]
                i += 1
            elif args[i] == "-l":
                logfile = args[i+1]
                i += 1
            elif args[i] == "-v":
                flags["verbose"] = True
            else:
                sys.stderr.write("Invalid argument: %s" % args[i])
                print_help()
                exit(1)
            i += 1
    except Exception as e:
        print e
        print_help()
        exit(1)

    today = datetime.now()

    logfiles = [{"verbose": False, "stream": sys.stdout}]
    errorfiles = [sys.stderr]
    if logfile != None:
        logfile = open(logfile, "w")
        logfiles.append({"verbose": True, "stream": logfile})
        errorfiles.append(logfile)


    t = Test()
    input_params = {}
    t.inputs = read_file(inputfile, 0, t, input_params)

    if output == None:
        output = "results.db"

    log_write(logfiles, "Writing to database %s..." % output)

    # Run the tests
    results = run_tests(t, logfiles, errorfiles)

    db = DB(output, logfiles)
    db.delete_old_results(t, results)

    # Write results to database
    for rs in results:
        db.insert_resultset(t, rs)


if __name__ == "__main__":
    main(sys.argv[1:])
