import matplotlib.pyplot as plt
import sys
import os
import numpy as np
import sqlite3
import re
import glob
from math import sqrt
import datetime

from plot_configurations import *

results = {}
DB_INPUT = None
YOKO_INPUT = None
OUTPUT_DIR = None
HALLGEIR_CONST = 0.001
YOKO_CONST = 0.01

bar_colors = [ 'b', 'r', 'g', 'c', 'm', 'y', 'w', 'k']

# Return conn and cursor
def connect_to_db(dbname):
    conn = sqlite3.connect(dbname)
    return (conn,conn.cursor())

# Create dictionary structure
def fill_dbresults(db_data):
    # row[0] = matrix, row[1] = mt_conf, row[2] = xform_conf, row[3] = enable_stat_preprocessing,
    # row[4] = cg_execution_time_mean, row[5] = cg_execution_time_stddev, row[6] = energy_pkg_mean, row[7] = energy_pkg_stddev,
    # row[8] = papi_l2_refs_mean, row[9] = papi_l2_refs_stddev, row[10] = papi_l2_miss_mean, row[11] = papi_l2_miss_stddev,
    # row[12] = papi_l3_refs_mean, row[13] = papi_l3_refs_stddev, row[15] = papi_l3_miss_mean, row[16] = papi_l3_miss_stddev,
    for row in db_data:
        # remove path
        matrix = re.sub('\.\/csx\/matrices\/\S+\/', '', row[0])
        # remove .mtx.sorted
        matrix = matrix[:-11]
        if matrix not in results.keys():
            results[matrix] = {}
        # here be dragons
        xform = "%s_st-%s" % (row[2], row[3])
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]] = {}
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['cg_execution_time_mean'] = row[4]
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['cg_execution_time_stddev'] = row[5]
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['energy_pkg_mean'] = row[6] * HALLGEIR_CONST
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['energy_pkg_stddev'] = row[7] * HALLGEIR_CONST
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['l2_refs_mean'] = row[8]
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['l2_refs_stddev'] = row[9]
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['l2_miss_mean'] = row[10]
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['l2_miss_stddev'] = row[11]
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['l3_refs_mean'] = row[12]
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['l3_refs_stddev'] = row[13]
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['l3_miss_mean'] = row[14]
        results[matrix][MTCONF_LUT[row[1]]+XFORM_LUT[xform]]['l3_miss_stddev'] = row[15]

def fill_yokoresults(yoko_dir):
    yoko_files = glob.glob("%s/*energy.dat" % yoko_dir)
    for yfile in yoko_files:
        stripped_filename = re.sub("^(.+/)+", "", yfile)
        m = re.match("(.+)_mt-([\d,]+)_xf-([\d,]+_st-\d)_energy.dat", stripped_filename)
        matrix = m.group(1)
        mt_conf = m.group(2)
        xform_conf = m.group(3)
        f = open(yfile, 'r')
        energy_readings = f.read().strip('\n').split('\t')
        f.close()
        mean = 0
        for num in energy_readings:
            mean += int(num)
        mean = mean / len(energy_readings)
        stddev = 0
        for num in energy_readings:
            stddev += pow((int(num)-mean), 2)
        stddev = sqrt(stddev / len(energy_readings))
        if matrix not in results.keys():
            results[matrix] = {}
        if MTCONF_LUT[mt_conf]+XFORM_LUT[xform_conf] not in results[matrix].keys(): 
            results[matrix][MTCONF_LUT[mt_conf]+XFORM_LUT[xform_conf]] = {}
        results[matrix][MTCONF_LUT[mt_conf]+XFORM_LUT[xform_conf]]['yoko_energy_mean'] = mean * YOKO_CONST
        results[matrix][MTCONF_LUT[mt_conf]+XFORM_LUT[xform_conf]]['yoko_energy_stddev'] = stddev * YOKO_CONST

def create_time_plot(graph_name, confs = None):
    if confs is None:
        confs = CONF_SORT_LUT
    groups = [
        ('all', MATRIX_LIST),
        ('low_nnz_group', ['LF10000', 'nasa2146', 'nasa2910', 'nasa4704', 'sts4098']),
        ('middle_group', ['2cubes_sphere', 'offshore', 'parabolic_fem', 'thermal1', 'thermal2']),
        ('belt_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10', 'gyro', 'nasasrb', 'olafu', 'Pres_Poisson', 'raefsky4', 'smt']),
        ('csx_optimized_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10']),
        ('semi_optimized_group', ['gyro', 'LF10000', 'nasasrb', 'offshore', 'parabolic_fem', 'smt', 'sts4098']),
        ('csr_optimized_group', ['2cubes_sphere', 'nasa2146', 'nasa2910', 'nasa4704', 'olafu', 'Pres_Poisson', 'raefsky4', 'thermal1', 'thermal2'])
    ]
    data = []
    index = 0
    for (group_name,matrices) in groups:
        newgroup = True
        for conf in confs:
            time = 0
            for matrix in matrices:
                time = time + results[matrix][conf]['cg_execution_time_mean']
            time = time / len(matrices)
            if newgroup:
                data.append((group_name, []))
                newgroup = False
            data[index][1].append(time)
        index = index + 1
    plt.figure()
    ind = np.arange(len(confs))
    for (group_name,values) in data:
        plt.plot(ind, values, label=group_name)
    plt.ylabel('Average runtime (s)')
#    plt.yscale('log')
    plt.xticks(ind, confs, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc='best', prop={'size':10})
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

def create_energy_plot(graph_name, confs = None):
    if confs is None:
        confs = CONF_SORT_LUT
    groups = [
        ('all', MATRIX_LIST),
        ('low_nnz_group', ['LF10000', 'nasa2146', 'nasa2910', 'nasa4704', 'sts4098']),
        ('middle_group', ['2cubes_sphere', 'offshore', 'parabolic_fem', 'thermal1', 'thermal2']),
        ('belt_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10', 'gyro', 'nasasrb', 'olafu', 'Pres_Poisson', 'raefsky4', 'smt']),
        ('csx_optimized_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10']),
        ('semi_optimized_group', ['gyro', 'LF10000', 'nasasrb', 'offshore', 'parabolic_fem', 'smt', 'sts4098']),
        ('csr_optimized_group', ['2cubes_sphere', 'nasa2146', 'nasa2910', 'nasa4704', 'olafu', 'Pres_Poisson', 'raefsky4', 'thermal1', 'thermal2'])
    ]
    data = []
    index = 0
    for (group_name,matrices) in groups:
        newgroup = True
        for conf in confs:
            energy = 0
            for matrix in matrices:
                energy = energy + results[matrix][conf]['energy_pkg_mean']
            energy = energy / len(matrices)
            if newgroup:
                data.append((group_name, []))
                newgroup = False
            data[index][1].append(energy)
        index = index + 1
    plt.figure()
    ind = np.arange(len(confs))
    for (group_name,values) in data:
        plt.plot(ind, values, label=group_name)
    plt.ylabel('Average energy consumption (J)')
#    plt.yscale('log')
    plt.xticks(ind, confs, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc='best', prop={'size':10})
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()



def create_time_plot_multiformat(graph_name, confs = None):
    if confs is None:
        confs = CONF_SORT_LUT
    groups = [
        ('all', MATRIX_LIST, ["R", "X"]),
        ('stat_prep', MATRIX_LIST_CSXS, ["R", "X", "Xs"]),
    ]
    data = []
    index = 0
    for (group_name,matrices, formats) in groups:
        for frm in formats:
            newgroup = True
            for conf in confs:
                time = 0
                for matrix in matrices:
                    time = time + results[matrix][conf+frm]['cg_execution_time_mean']
                time = time / len(matrices)
                if newgroup:
                    data.append((group_name+"_CS"+frm, []))
                    newgroup = False
                data[index][1].append(time)
            index = index + 1
    plt.figure()
    ind = np.arange(len(confs))
    for (group_name,values) in data:
        plt.plot(ind, values, label=group_name)
    plt.ylabel('Average runtime (s)')
#    plt.yscale('log')
    plt.xticks(ind, confs, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc='best', prop={'size':10})
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

def create_energy_plot_multiformat(graph_name, confs = None):
    if confs is None:
        confs = CONF_SORT_LUT
    groups = [
        ('all', MATRIX_LIST, ["R", "X"]),
        ('stat_prep', MATRIX_LIST_CSXS, ["R", "X", "Xs"]),
    ]
    data = []
    index = 0
    for (group_name,matrices, formats) in groups:
        for frm in formats:
            newgroup = True
            for conf in confs:
                energy = 0
                for matrix in matrices:
                    energy = energy + results[matrix][conf+frm]['energy_pkg_mean']
                energy = energy / len(matrices)
                if newgroup:
                    data.append((group_name+"_CS"+frm, []))
                    newgroup = False
                data[index][1].append(energy)
            index = index + 1
    plt.figure()
    ind = np.arange(len(confs))
    for (group_name,values) in data:
        plt.plot(ind, values, label=group_name)
    plt.ylabel('Average energy consumption (J)')
#    plt.yscale('log')
    plt.xticks(ind, confs, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc='best', prop={'size':10})
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

# Energy per nnz value
def create_eng_per_nnz_plot(graph_name, confs = None):
    if confs is None:
        confs = CONF_SORT_LUT
    data = {}
# Middle matrices
#    skip_matrices = ['LFAT5', 'nasa2146', 'nasa2910', 'nasa4704', 'sts4098', 'LF10000', 'af_5_k101', 'boneS01', 'smt', 'Pres_Poisson', 'nasasrb', 'raefsky4', 'olafu', 'boneS10', 'gyro', 'bone010', 'af_shell3']
# Good matrices
    skip_matrices = ['LFAT5', 'nasa2146', 'nasa2910', 'nasa4704', 'sts4098', 'LF10000', 'thermal2', 'thermal1', 'parabolic_fem', 'offshore', '2cubes_sphere']
# Low nnz matrices
#    skip_matrices = ['LFAT5', 'af_5_k101', 'boneS01', 'smt', 'Pres_Poisson', 'nasasrb', 'raefsky4', 'olafu', 'boneS10', 'gyro', 'bone010', 'af_shell3', 'thermal2', 'thermal1', 'parabolic_fem', 'offshore', '2cubes_sphere']
# All
#    skip_matrices = ['LFAT5']
    for conf in confs:
        for matrix in results.keys():
            if matrix in skip_matrices:
                continue
            if matrix not in MATRIX_DATA.keys():
                print "Matrix %s is not defined in configuration!" % matrix
                exit(1)
            if matrix not in data.keys():
                data[matrix] = []
            data[matrix].append(results[matrix][conf]['energy_pkg_mean']/MATRIX_DATA[matrix]['nnz'])
    plt.figure()
    ind = np.arange(len(confs))
    mat_count = 0
    for matrix in results.keys():
        if matrix in skip_matrices:
            continue
        if mat_count > 13:
            plt.plot(ind, data[matrix], ls=':', label=matrix)
        elif mat_count > 6:
            plt.plot(ind, data[matrix], ls='--', label=matrix)
        else:
            plt.plot(ind, data[matrix], label=matrix)
        mat_count += 1
    plt.ylabel('energy (J)/nnz')
    plt.xticks(ind, confs, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc=9, prop={'size':10}, ncol=3, bbox_to_anchor=(0.5,1.12))
    plt.ylim(0.00000,0.00016)
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

# Energy decrease for each matrix
def create_eng_dec_plot(graph_name, confs = None):
    if confs is None:
        confs = CONF_SPEEDUP_MIN_SORT_LUT
    data = []
# All
    matrices = MATRIX_LIST
# Low nnz matrices
#    matrices = ['LF10000','nasa2146', 'nasa2910', 'nasa4704', 'sts4098']
# Middle matrices
#    matrices = ['2cubes_sphere', 'offshore', 'parabolic_fem', 'thermal1', 'termal2'
# Belt matrices
#    matrices = ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10', 'gyro', 'nasasrb', 'olafu', 'Pres_Poisson', 'raefsky4', 'smt']
    index = 0
    for matrix in matrices:
        new_matrix = True
        for conf in confs:
            if matrix not in MATRIX_DATA.keys():
                print "Matrix %s is not defined in configuration!" % matrix
                exit(1)
            if new_matrix:
                data.append((matrix,[]))
                new_matrix = False
            data[index][1].append(results[matrix][conf+"R"]['energy_pkg_mean']/results[matrix][conf+"X"]['energy_pkg_mean'])
        index = index + 1
    plt.figure()
    ind = np.arange(len(confs))
    mat_count = 0
    for (matrix,values) in data:
        if mat_count > 13:
            plt.plot(ind, values, ls=':', label=matrix)
        elif mat_count > 6:
            plt.plot(ind, values, ls='--', label=matrix)
        else:
            plt.plot(ind, values, label=matrix)
        mat_count = mat_count + 1
    plt.ylabel('Energy decrease ratio')
    plt.xticks(ind, confs, rotation=30, fontsize='10')
    plt.legend(loc=9, prop={'size':10}, ncol=3, bbox_to_anchor=(0.5,1.12))
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

# Average energy decrease for groups of matrices
def create_group_eng_dec_plot(graph_name, confs = None):
    if confs is None:
        confs = CONF_SPEEDUP_MIN_SORT_LUT
    # Groups of matrices that will have their average edr shown in plot
    groups = [
        ('all', MATRIX_LIST),
        ('low_nnz_group', ['LF10000', 'nasa2146', 'nasa2910', 'nasa4704', 'sts4098']),
        ('middle_group', ['2cubes_sphere', 'offshore', 'parabolic_fem', 'thermal1', 'thermal2']),
        ('belt_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10', 'gyro', 'nasasrb', 'olafu', 'Pres_Poisson', 'raefsky4', 'smt']),
        ('csx_optimized_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10']),
        ('semi_optimized_group', ['gyro', 'LF10000', 'nasasrb', 'offshore', 'parabolic_fem', 'smt', 'sts4098']),
        ('csr_optimized_group', ['2cubes_sphere', 'nasa2146', 'nasa2910', 'nasa4704', 'olafu', 'Pres_Poisson', 'raefsky4', 'thermal1', 'thermal2'])
    ]
    data = []
    index = 0
    for (group_name,matrices) in groups:
        newgroup = True
        for conf in confs:
            speedup = 0
            for matrix in matrices:
                speedup = speedup + results[matrix][conf+"R"]['energy_pkg_mean']/results[matrix][conf+"X"]['energy_pkg_mean']
            speedup = speedup / len(matrices)
            if newgroup:
                data.append((group_name, []))
                newgroup = False
            data[index][1].append(speedup)
        index = index + 1
    plt.figure()
    ind = np.arange(len(confs))
    for (group_name,values) in data:
        plt.plot(ind, values, label=group_name)
    plt.ylabel('Energy decrease ratio')
    plt.xticks(ind, confs, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc='best', prop={'size':10})
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

def create_group_miss_rate_plot(graph_name, confs = None):
    if confs is None:
        confs = CONF_SPEEDUP_MIN_SORT_LUT
    groups = [
        ('all', results.keys()),
        ('low_nnz_group', ['LF10000', 'nasa2146', 'nasa2910', 'nasa4704', 'sts4098']),
        ('bad_group', ['2cubes_sphere', 'offshore', 'parabolic_fem', 'thermal1', 'thermal2']),
        ('belt_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10', 'gyro', 'nasasrb', 'olafu', 'Pres_Poisson', 'raefsky4', 'smt']),
        ('csx_optimized_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10']),
        ('semi_optimized_group', ['gyro', 'LF10000', 'nasasrb', 'offshore', 'parabolic_fem', 'smt', 'sts4098']),
        ('csr_optimized_group', ['2cubes_sphere', 'nasa2146', 'nasa2910', 'nasa4704', 'olafu', 'Pres_Poisson', 'raefsky4', 'thermal1', 'thermal2'])
    ]
    cs_types = ['R', 'X']
    miss = 'l2'
    data = []
    index = 0
    for (group_name,matrices) in groups:
        for cs_type in cs_types:
            newgroup = True
            for conf in confs:
                miss_rate = 0;
                for matrix in matrices:
                    miss_rate = miss_rate + results[matrix][conf+cs_type][miss+'_miss_mean']/results[matrix][conf+cs_type][miss+'_refs_mean']
                miss_rate = (miss_rate / len(matrices)) * 100
                if newgroup:
                    data.append((group_name+'_CS'+cs_type+'_'+miss, []))
                    newgroup = False
                data[index][1].append(miss_rate)
            index = index + 1
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ind = np.arange(len(confs))
    plot_num = 0
    for (group_name,values) in data:
        ls = ''
        if plot_num % 2 == 0:
            line = '-'
        else:
            line = '--'
        if plot_num > 11:
            ax.plot(ind, values, ls=line, label=group_name, c='k')
        elif plot_num > 9:
            ax.plot(ind, values, ls=line, label=group_name, c='y')
        elif plot_num > 7:
            ax.plot(ind, values, ls=line, label=group_name, c='m')
        elif plot_num > 5:
            ax.plot(ind, values, ls=line, label=group_name, c='c')
        elif plot_num > 3:
            ax.plot(ind, values, ls=line, label=group_name, c='r')
        elif plot_num > 1:
            ax.plot(ind, values, ls=line, label=group_name, c='g')
        else:
            ax.plot(ind, values, ls=line, label=group_name, c='b')
        plot_num += 1
    plt.ylabel('Miss rates (%)')
    plt.xticks(ind, confs, rotation=30, fontsize='10')
    plt.grid(which='both')
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.65, box.height])
    ax.legend(loc=7, prop={'size':10}, bbox_to_anchor=(1.7,0.5))
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

def create_group_socket_energy_comp_plot(graph_name, confs = None):
    if confs is None:
        confs = CONF_SOCKET_CSR_SORT_LUT
    # Groups of matrices that will have their average edr shown in plot
    groups = [
        ('all', MATRIX_LIST),
        ('low_nnz_group', ['LF10000', 'nasa2146', 'nasa2910', 'nasa4704', 'sts4098']),
        ('middle_group', ['2cubes_sphere', 'offshore', 'parabolic_fem', 'thermal1', 'thermal2']),
        ('belt_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10', 'gyro', 'nasasrb', 'olafu', 'Pres_Poisson', 'raefsky4', 'smt']),
        ('csx_optimized_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10']),
        ('semi_optimized_group', ['gyro', 'LF10000', 'nasasrb', 'offshore', 'parabolic_fem', 'smt', 'sts4098']),
        ('csr_optimized_group', ['2cubes_sphere', 'nasa2146', 'nasa2910', 'nasa4704', 'olafu', 'Pres_Poisson', 'raefsky4', 'thermal1', 'thermal2'])
    ]
    data = []
    index = 0
    for (group_name,matrices) in groups:
        newgroup = True
        i = 0
        while i < len(confs):
            energy_ratio = 0
            for matrix in matrices:
                energy_ratio = energy_ratio + results[matrix][confs[i+1]]['energy_pkg_mean'] - results[matrix][confs[i]]['energy_pkg_mean']
            energy_ratio = energy_ratio / len(matrices)
            if newgroup:
                data.append((group_name, []))
                newgroup = False
            data[index][1].append(energy_ratio)
            i += 2
        index = index + 1
    plt.figure()
    ind = np.arange(len(confs)/2)
    conf_names = ['2t0s', '4t0s', '8t0s']
    for (group_name,values) in data:
        plt.plot(ind, values, label=group_name)
    plt.ylabel('Energy difference (J)')
    plt.xticks(ind, conf_names, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc='best', prop={'size':10})
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

def create_group_socket_speedup_plot(graph_name, confs = None):
    if confs is None:
        confs = CONF_SOCKET_CSR_SORT_LUT
    # Groups of matrices that will have their average edr shown in plot
    groups = [
        ('all', MATRIX_LIST),
        ('low_nnz_group', ['LF10000', 'nasa2146', 'nasa2910', 'nasa4704', 'sts4098']),
        ('middle_group', ['2cubes_sphere', 'offshore', 'parabolic_fem', 'thermal1', 'thermal2']),
        ('belt_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10', 'gyro', 'nasasrb', 'olafu', 'Pres_Poisson', 'raefsky4', 'smt']),
        ('csx_optimized_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10']),
        ('semi_optimized_group', ['gyro', 'LF10000', 'nasasrb', 'offshore', 'parabolic_fem', 'smt', 'sts4098']),
        ('csr_optimized_group', ['2cubes_sphere', 'nasa2146', 'nasa2910', 'nasa4704', 'olafu', 'Pres_Poisson', 'raefsky4', 'thermal1', 'thermal2'])
    ]
    data = []
    index = 0
    for (group_name,matrices) in groups:
        newgroup = True
        i = 0
        while i < len(confs):
            speedup = 0
            for matrix in matrices:
                speedup = speedup + results[matrix][confs[i]]['cg_execution_time_mean']/results[matrix][confs[i+1]]['cg_execution_time_mean']
            speedup = speedup / len(matrices)
            if newgroup:
                data.append((group_name, []))
                newgroup = False
            data[index][1].append(speedup)
            i += 2
        index = index + 1
    plt.figure()
    ind = np.arange(len(confs)/2)
    conf_names = ['2t0s', '4t0s', '8t0s']
    for (group_name,values) in data:
        plt.plot(ind, values, label=group_name)
    plt.ylabel('Speedup')
    plt.xticks(ind, conf_names, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc='best', prop={'size':10})
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

def create_group_miss_rate_socket_plot(graph_name, confs = None):
    if confs is None:
        confs = CONF_SOCKET_CSR_SORT_LUT
    groups = [
        ('all', MATRIX_LIST),
        ('low_nnz_group', ['LF10000', 'nasa2146', 'nasa2910', 'nasa4704', 'sts4098']),
        ('bad_group', ['2cubes_sphere', 'offshore', 'parabolic_fem', 'thermal1', 'thermal2']),
        ('belt_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10', 'gyro', 'nasasrb', 'olafu', 'Pres_Poisson', 'raefsky4', 'smt']),
        ('csx_optimized_group', ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10']),
        ('semi_optimized_group', ['gyro', 'LF10000', 'nasasrb', 'offshore', 'parabolic_fem', 'smt', 'sts4098']),
        ('csr_optimized_group', ['2cubes_sphere', 'nasa2146', 'nasa2910', 'nasa4704', 'olafu', 'Pres_Poisson', 'raefsky4', 'thermal1', 'thermal2'])
    ]
    miss = 'l3'
    data = []
    index = 0
    for (group_name,matrices) in groups:
        newgroup = True
        i = 0
        while i < len(confs):
            miss_rate_1c = 0;
            miss_rate_2c = 0;
            for matrix in matrices:
                miss_rate_1c = miss_rate_1c + results[matrix][confs[i]][miss+'_miss_mean']/results[matrix][confs[i]][miss+'_refs_mean']
                miss_rate_2c = miss_rate_2c + results[matrix][confs[i+1]][miss+'_miss_mean']/results[matrix][confs[i+1]][miss+'_refs_mean']
            miss_rate_1c = (miss_rate_1c / len(matrices)) * 100
            miss_rate_2c = (miss_rate_2c / len(matrices)) * 100
            if newgroup:
                data.append((group_name+'_1C_'+miss, []))
                data.append((group_name+'_2C_'+miss, []))
                newgroup = False
            data[index][1].append(miss_rate_1c)
            data[index+1][1].append(miss_rate_2c)
            i += 2
        index = index + 2
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ind = np.arange(len(confs)/2)
    plot_num = 0
    for (group_name,values) in data:
        ls = ''
        if plot_num % 2 == 0:
            line = '-'
        else:
            line = '--'
        if plot_num > 11:
            ax.plot(ind, values, ls=line, label=group_name, c='k')
        elif plot_num > 9:
            ax.plot(ind, values, ls=line, label=group_name, c='y')
        elif plot_num > 7:
            ax.plot(ind, values, ls=line, label=group_name, c='m')
        elif plot_num > 5:
            ax.plot(ind, values, ls=line, label=group_name, c='c')
        elif plot_num > 3:
            ax.plot(ind, values, ls=line, label=group_name, c='r')
        elif plot_num > 1:
            ax.plot(ind, values, ls=line, label=group_name, c='g')
        else:
            ax.plot(ind, values, ls=line, label=group_name, c='b')
        plot_num += 1
    plt.ylabel('Miss rates (%)')
    if confs == CONF_SOCKET_CSR_SORT_LUT:
        conf_names = ['2t0sR', '4t0sR', '8t0sR']
    elif confs == CONF_SOCKET_CSX_SORT_LUT:
        conf_names = ['2t0sX', '4t0sX', '8t0sX']
    else:
        conf_names = ['2t0s', '4t0s', '8t0s']
    plt.xticks(ind, conf_names, rotation=30, fontsize='10')
    plt.grid(which='both')
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.65, box.height])
    ax.legend(loc=7, prop={'size':10}, bbox_to_anchor=(1.7,0.5))
    plt.ylim(0,30)
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

def create_miss_per_nnz_plot(graph_name, confs = None):
    if confs is None:
        confs = CONF_SORT_LUT
    data = {}
    threads = 1
    for conf in confs:
        print "%s: %d" % (conf, threads)
        for matrix in MATRIX_LIST:
            if matrix not in MATRIX_DATA.keys():
                print "Matrix %s is not defined in configuration!" % matrix
                exit(1)
            if matrix not in data.keys():
                data[matrix] = []
#            data[matrix].append((results[matrix][conf]['l3_miss_mean'] * threads)/MATRIX_DATA[matrix]['nnz'])
            data[matrix].append((results[matrix][conf]['l3_miss_mean'] * threads)/results[matrix][conf]['cg_execution_time_mean'])
        threads *= 2
    plt.figure()
    ind = np.arange(len(confs))
    mat_count = 0
    for matrix in MATRIX_LIST:
        if mat_count > 13:
            plt.plot(ind, data[matrix], ls=':', label=matrix)
        elif mat_count > 6:
            plt.plot(ind, data[matrix], ls='--', label=matrix)
        else:
            plt.plot(ind, data[matrix], label=matrix)
        mat_count += 1
    plt.ylabel('L3 misses/second')
    plt.xticks(ind, confs, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc=9, prop={'size':10}, ncol=3, bbox_to_anchor=(0.5,1.12))
    plt.ylim(0,1.2e8)
#    plt.yscale('log')
    plt.savefig("%s/%s.png" % (OUTPUT_DIR, graph_name))
    plt.close()

def create_msr_best_conf_plot(graph_name):
    data_eng = {}
    data_perf = {}
    check_mat = ['parabolic_fem', 'boneS10', 'bone010']
    for matrix in MATRIX_LIST:
        best_eng = 9e99
        best_eng_conf = ''
        best_perf = 9e99
        best_perf_conf = ''
        for conf in results[matrix].keys():
            if results[matrix][conf]['energy_pkg_mean'] < best_eng:
                best_eng = results[matrix][conf]['energy_pkg_mean']
                best_eng_conf = conf
            if results[matrix][conf]['cg_execution_time_mean'] < best_perf:
                best_perf = results[matrix][conf]['cg_execution_time_mean']
                best_perf_conf = conf
            if matrix in check_mat:
                print "%s, %s: %f, %f" % (matrix, conf, results[matrix][conf]['energy_pkg_mean'], results[matrix][conf]['cg_execution_time_mean'])
        if best_eng_conf not in data_eng.keys():
            data_eng[best_eng_conf] = [matrix]
        else:
            data_eng[best_eng_conf].append(matrix)
        if best_perf_conf not in data_perf.keys():
            data_perf[best_perf_conf] = [matrix]
        else:
            data_perf[best_perf_conf].append(matrix)
    print data_eng
    print data_perf
    groups = [
        ['LF10000', 'nasa2146', 'nasa2910', 'nasa4704', 'sts4098'],
        ['2cubes_sphere', 'offshore', 'parabolic_fem', 'thermal1', 'thermal2'],
        ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10', 'gyro', 'nasasrb', 'olafu', 'Pres_Poisson', 'raefsky4', 'smt'],
        ['2cubes_sphere', 'nasa2146', 'nasa2910', 'nasa4704', 'olafu', 'Pres_Poisson', 'raefsky4', 'thermal1', 'thermal2'],
        ['gyro', 'LF10000', 'nasasrb', 'offshore', 'parabolic_fem', 'smt', 'sts4098'],
        ['af_5_k101', 'af_shell3', 'bone010', 'boneS01', 'boneS10']
    ]
    # Energy plot
    plt.figure()
    ind = np.arange(len(data_eng.keys()))
    width = 0.35
    bars = []
    xticks = []
    group_numbers = [[0 for x in range(len(data_eng.keys()))] for x in range(6)]
    index = 0
    for conf in CONF_SORT_LUT:
        if conf in data_eng.keys():
            for matrix in data_eng[conf]:
                for i in range(6):
                    if matrix in groups[i]:
                        group_numbers[i][index] += 1
            xticks.append(conf)
            index += 1
    plt.bar(ind, group_numbers[0], width, label='low_nnz_group', color='b')
    plt.bar(ind, group_numbers[1], width, bottom=group_numbers[0], label='middle_group', color='r')
    bottom = [sum(a) for a in zip(group_numbers[0], group_numbers[1])]
    plt.bar(ind, group_numbers[2], width, bottom=bottom, label='belt_group', color='g')

    plt.bar(ind+width, group_numbers[3], width, label='csr_optimized_group', color='c')
    plt.bar(ind+width, group_numbers[4], width, bottom=group_numbers[3], label='semi_optimized_group', color='m')
    bottom = [sum(a) for a in zip(group_numbers[3], group_numbers[4])]
    plt.bar(ind+width, group_numbers[5], width, bottom=bottom, label='csx_optimized_group', color='y')

    plt.ylabel('#matrices optimized by configuration')
    plt.xticks(ind+width, xticks, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc='best', prop={'size':10})
    plt.savefig("%s/%s_eng.png" % (OUTPUT_DIR, graph_name))
    plt.close()

    # Runtime plot
    plt.figure()
    ind = np.arange(len(data_perf.keys()))
    width = 0.35
    bars = []
    xticks = []
    group_numbers = [[0 for x in range(len(data_perf.keys()))] for x in range(6)]
    index = 0
    for conf in CONF_SORT_LUT:
        if conf in data_perf.keys():
            for matrix in data_perf[conf]:
                for i in range(6):
                    if matrix in groups[i]:
                        group_numbers[i][index] += 1
            xticks.append(conf)
            index += 1
    plt.bar(ind, group_numbers[0], width, label='low_nnz_group', color='b')
    plt.bar(ind, group_numbers[1], width, bottom=group_numbers[0], label='middle_group', color='r')
    bottom = [sum(a) for a in zip(group_numbers[0], group_numbers[1])]
    plt.bar(ind, group_numbers[2], width, bottom=bottom, label='belt_group', color='g')

    plt.bar(ind+width, group_numbers[3], width, label='csr_optimized_group', color='c')
    plt.bar(ind+width, group_numbers[4], width, bottom=group_numbers[3], label='semi_optimized_group', color='m')
    bottom = [sum(a) for a in zip(group_numbers[3], group_numbers[4])]
    plt.bar(ind+width, group_numbers[5], width, bottom=bottom, label='csx_optimized_group', color='y')

    plt.ylabel('#matrices optimized by configuration')
    plt.xticks(ind+width, xticks, rotation=30, fontsize='10')
    plt.grid(which='both')
    plt.legend(loc='best', prop={'size':10})
    plt.savefig("%s/%s_perf.png" % (OUTPUT_DIR, graph_name))
    plt.close()

for i in range(1,len(sys.argv),2):
    if (sys.argv[i] == "-db"):
        DB_INPUT = sys.argv[i+1]
    elif (sys.argv[i] == "-yd"):
        YOKO_INPUT = sys.argv[i+1]
    elif (sys.argv[i] == "-out"):
        OUTPUT_DIR = sys.argv[i+1]
    else:
        print "Invalid argument: %s, usage: %s [-db dbname] [-yd yokogawa_dir] -out output_dir" % (sys.argv[i], sys.argv[0])
        sys.exit(1)

if not (DB_INPUT or YOKO_INPUT):
    print "No input specified! usage: %s [-db powertime_dir] [-yd yokogawa_dir] -out output_dir" % sys.argv[0]
    sys.exit(1)

if not OUTPUT_DIR:
    print "No output directory specified! usage: %s [-db powertime_dir] [-yd yokogawa_dir] -out graph_dir" % sys.argv[0]
    sys.exit(1)

if DB_INPUT and not (os.path.exists(DB_INPUT) and os.path.isfile(DB_INPUT)):
    print "Database does not exist or is not a file: %s" % DB_INPUT
    sys.exit(1)

if YOKO_INPUT and not (os.path.exists(YOKO_INPUT) and os.path.isdir(YOKO_INPUT)):
    print "Yokogawa directory does not exist or is not a directory: %s" % YOKO_INPUT
    sys.exit(1)

if YOKO_INPUT and not glob.glob("%s/*energy.dat" % YOKO_INPUT):
    print "Cannot find any energy .dat files in %s, exiting" % YOKO_INPUT
    sys.exit(1)

if not os.path.exists(OUTPUT_DIR):
    os.makedirs(OUTPUT_DIR)

if (os.path.exists(OUTPUT_DIR) and not os.path.isdir(OUTPUT_DIR)):
    print "Output directory is not a directory: %s" % OUTPUT_DIR
    sys.exit(1)

if DB_INPUT:
    (conn,c) = connect_to_db(DB_INPUT)
    query = """
        SELECT
            matrix, mt_conf, xform_conf, enable_stat_preprocessing,
            cg_execution_time_mean, cg_execution_time_stddev, energy_pkg_mean, energy_pkg_stddev,
            papi_l2_refs_mean, papi_l2_refs_stddev, papi_l2_miss_mean, papi_l2_miss_stddev,
            papi_l3_refs_mean, papi_l3_refs_stddev, papi_l3_miss_mean, papi_l3_miss_stddev
        FROM
            results
        ORDER BY
            matrix;"""
    fill_dbresults(c.execute(query))
    conn.close()

if YOKO_INPUT:
    fill_yokoresults(YOKO_INPUT)

#create_eng_dec_plot("energy_decrease")
#create_eng_per_nnz_plot("csr_plot", CONF_CSR_MIN_SORT_LUT)
#create_eng_per_nnz_plot("csx_plot", CONF_CSX_MIN_SORT_LUT)
#create_group_eng_dec_plot("energy_decrease", CONF_PARA_COMP_SORT_LUT)
#create_group_socket_energy_comp_plot("socket_csr", CONF_SOCKET_CSR_SORT_LUT)
#create_group_socket_energy_comp_plot("socket_csx", CONF_SOCKET_CSX_SORT_LUT)
#create_group_socket_speedup_plot("socket_csr", CONF_SOCKET_CSR_SORT_LUT)
#create_group_socket_speedup_plot("socket_csx", CONF_SOCKET_CSX_SORT_LUT)
#create_group_miss_rate_socket_plot("misses_csr", CONF_SOCKET_CSR_SORT_LUT )
#create_group_miss_rate_socket_plot("misses_csx", CONF_SOCKET_CSX_SORT_LUT )
#create_time_plot_multiformat("timeplot", CONF_SPEEDUP_MIN_SORT_LUT);
#create_energy_plot_multiformat("energy", CONF_SPEEDUP_MIN_SORT_LUT);
#create_time_plot("timeplot_csr", CONF_PARA_CSR_SORT_LUT);
#create_time_plot("timeplot_csx", CONF_PARA_CSX_SORT_LUT);
#create_energy_plot("energy_csr", CONF_PARA_CSR_SORT_LUT);
#create_energy_plot("energy_csx", CONF_PARA_CSX_SORT_LUT);
#create_miss_per_nnz_plot("miss_per_nnz_csr", CONF_CSR_MIN_SORT_LUT)
#create_miss_per_nnz_plot("miss_per_nnz_csx", CONF_CSX_MIN_SORT_LUT)
create_msr_best_conf_plot("msr_best_conf")

sys.exit()
